from ._core_base import _BaseObject, _BaseProperty, ChildError, ksplit, BaseData, kjoin, _get_class_dict, open_object
from ._core_device import _DeviceListing, BaseDevice, parse_device_map
from ._core_node import BaseNode, parse_node_map, _NodeListing
from ._core_rpc import BaseRpc, parse_rpc_map, _RpcListing
from ._core_interface import BaseInterface, parse_interface_map, _InterfaceListing
from ._class_recorder import KINDS, get_class, record_class

from pydantic import BaseModel, validator
from typing import Dict, Union

class ManagerProperty(_BaseProperty):    
    fbuild = None    
    def builder(self, func):
        """ Decorator for the interface builder """
        self.fbuild = func
        return self
    
    def __call__(self, func):
        """ The call is used has fget decorator """
        self.fbuild = func
        return self
    
    def _finalise(self, parent, device):
        # overwrite the fget, fset function to the node if defined         
        if self.fbuild:
            self.fbuild(parent, device)  
                      
class ManagerConfig(_BaseObject.Config):
    kind: KINDS = KINDS.MANAGER.value
    type: str = "Base"
    map_built: bool=  False
    map_built: bool = False
    
    interface_map : Dict[str, Union[BaseInterface.Config, dict]] = {} 
    node_map : Dict[str, Union[BaseNode.Config, dict]] = {}
    rpc_map : Dict[str, Union[BaseRpc.Config, dict]] = {} 
    device_map : Dict[str, Union[BaseDevice.Config, dict]] = {}
        
    @validator('device_map', pre=True, always=True)
    def _validate_device_map(cls, map, values):
        if values.get('map_built', False):
            return map
        cls = _get_class_dict(values)

        # cls = get_class(values['kind'], values['type'])
        if map is None:
            map = cls.default_device_map()
        cls.parse_device_map(map)
        return map
    
    @validator('rpc_map', pre=True, always=True)
    def _validate_rpc_map(cls, map, values):
        if values.get('map_built', False):
            return map
        cls = _get_class_dict(values)

        # cls = get_class(values['kind'], values['type'])

        if map is None:
            map = cls.default_rpc_map()
        cls.parse_rpc_map(map)
        return map
    
    @validator('node_map', pre=True, always=True)
    def _validate_node_map(cls, map, values):
        if values.get('map_built', False):
            return map
        cls = _get_class_dict(values)

        # cls = get_class(values['kind'], values['type'])   
        if map is None:
            map = cls.default_node_map()     
        cls.parse_node_map(map)
        return map
    
    @validator('interface_map', pre=True, always=True)
    def _validate_interface_map(cls, map, values):
        if values.get('map_built', False):
            return map        
        
        cls = _get_class_dict(values)

        # cls = get_class(values['kind'], values['type'])
        
        if map is None:
            map = cls.default_interface_map()
        cls.parse_interface_map(map)
        return map        

    def cfgdict(self, exclude=set()):     
        
        all_exclude = {*{"map_built", "node_map", "rpc_map", "interface_map","device_map"}, *exclude}
        d = super().cfgdict(exclude=all_exclude)
        for map  in ('node_map', 'rpc_map',  'interface_map','device_map'):
            if map not in exclude:
                d[map] = {k:n.cfgdict() for k,n in getattr(self, map).items()}
        return d


def open_manager(cfgfile, path=None, prefix="", key=None, default_type=None, **kwargs):
    """ Open a manager from a configuration file 

        
        Args:
            cfgfile: relative path to one of the $CFGPATH or absolute path to the yaml config file 
            key: Key of the created Manager 
            path (str, int, optional): 'a.b.c' will loock to cfg['a']['b']['c'] in the file. If int it will loock to the Nth
                                        element in the file
            prefix (str, optional): additional prefix added to the name or key

        Output:
            manager (BaseManager subclass) :tanciated Manager class     
    """
    kwargs.setdefault("kind", KINDS.MANAGER)
    return open_object(
                cfgfile, 
                path=path, prefix=prefix, 
                key=key, default_type=default_type, 
                **kwargs
            ) 





@record_class        
class BaseManager(_BaseObject, _DeviceListing, _NodeListing, _RpcListing, _InterfaceListing):
    Property  = ManagerProperty
    Config = ManagerConfig
    Data = BaseData
    Device = BaseDevice
    Interface = BaseInterface
    Node = BaseNode
    Rpc = BaseRpc
    
    def __init__(self, *args, devices=None, **kwargs):
        super().__init__(*args, **kwargs)
        if devices:
            for name, obj in devices.items():
                if isinstance(obj, BaseDevice):
                    self.__dict__[name] = obj
                    self.config.device_map[name] = obj.config
                elif isinstance(obj, BaseDevice.Config):
                    Device = get_class(obj.kind, obj.type)
                    self.__dict__[name]  = Device(kjoin( self.key, name), config=obj)
                    self.config.device_map[name] = obj
                elif isinstance(obj, dict):
                    try:
                        tpe = obj['type']
                    except KeyError:
                        raise ValueError('For device "{name}" the type is missing ')
                    try:
                        kind = obj['kind']
                    except KeyError:
                        raise ValueError('For device "{name}" the kind is missing ')
                    Device = get_class(kind, tpe)
                    config = Device.Config.from_cfgdict(obj)
                    self.__dict__[name]  = Device(kjoin( self.key, name), config=config)
                    self.config.device_map[name] = obj    
                
    def connect(self) -> None:
        """ Connect all devices """
        for device in self.devices:
            device.connect()
    
    def disconnect(self) -> None:
        """ disconnect all devices """
        for device in self.devices:
            device.disconnect()                
                
                    
    @classmethod
    def parse_config(cls, config, **kwargs):
        if isinstance(config, dict):
            kwargs = {**config, **kwargs}
            config = None
        if config is  None:
            kwargs['map_built'] = True
            parse_device_map(cls, kwargs.setdefault('device_map', {}))
            parse_interface_map(cls, kwargs.setdefault('interface_map', {}))
            parse_node_map(cls, kwargs.setdefault('node_map', {}))
            parse_rpc_map(cls, kwargs.setdefault('rpc_map', {}))
            
        return super().parse_config(config, **kwargs)
            
    
    def __getattr__(self, attr):
        try:
            return object.__getattribute__(self, attr)
        except AttributeError:
            try:
                return self._build_interface(attr)
            except ChildError:
                pass
            try:
                return self._build_node(attr)
            except ChildError:
                pass
            try:
                return self._build_rpc(attr)
            except ChildError:
                pass
            try:
                return self._build_device(attr)
            except ChildError:
                pass
            raise AttributeError(attr)            
        
    def clear(self):
        """ clear all cashed intermediate objects """
        self._clear_all()
    
        



from ._core_base import (_BaseObject, _BaseProperty, ObjectIterator, ChildError,
        IOConfig, ksplit, BaseData,
        open_object, _get_class_dict)
from ._class_recorder import get_device_class, KINDS, get_class, record_class
from ._core_node import BaseNode, _NodeListing, parse_node_map
from ._core_interface import BaseInterface, _InterfaceListing, parse_interface_map
from ._core_com import BaseCom
from ._core_rpc import BaseRpc, _RpcListing, parse_rpc_map
from pydantic  import BaseModel, validator
from .. import io 

from typing import List, Optional, Any, Dict, Union, Type





class BaseDeviceConfig(_BaseObject.Config):
    kind: KINDS = KINDS.DEVICE.value
    type: str = "Base"
    com: Optional[BaseCom.Config] = None
    
    map_built: bool = False
    
    interface_map : Dict[str, Union[BaseInterface.Config, dict]] = None
    node_map : Dict[str, Union[BaseNode.Config, dict]] = None
    rpc_map : Dict[str, Union[BaseRpc.Config, dict]] = None
    
    @validator('rpc_map', pre=True, always=True)
    def _validate_rpc_map(cls, map, values):
        if values.get('map_built', False):
            return map
        cls = _get_class_dict(values)
        # cls = get_class(values['kind'], values['type'])
        if map is None:
            map = cls.default_rpc_map()
            
        cls.parse_rpc_map(map)
        return map
    
    @validator('node_map', pre=True, always=True)
    def _validate_node_map(cls, map, values):
        if values.get('map_built', False):
            return map
        cls = _get_class_dict(values)
        # cls = get_class(values['kind'], values['type'])        
        if map is None:
            map = cls.default_node_map()
        cls.parse_node_map(map)
        return map
    
    @validator('interface_map', pre=True, always=True)
    def _validate_interface_map(cls, map, values):
        if values.get('map_built', False):
            return map        
        cls = _get_class_dict(values)
    
        # cls = get_class(values['kind'], values['type'])
        if map is None:
            map = cls.default_interface_map()
        cls.parse_interface_map(map)
        return map
    
    def cfgdict(self, exclude=set()):
        all_exclude = {*{"map_built", "node_map", "rpc_map", "interface_map","device_map"}, *exclude}
        d = super().cfgdict(exclude=all_exclude)
        for map  in ('node_map', 'rpc_map',  'interface_map','device_map'):
            if map not in exclude:
                d[map] = {k:n.cfgdict() for k,n in getattr(self, map).items()}        
        return d
    
    class Config:
        validate_assignment = True

class BaseDeviceConfig(BaseDeviceConfig):
    device_map : Dict[str, Union[BaseDeviceConfig, dict]] = {}
    
    @validator('device_map', pre=True, always=True)
    def _validate_device_map(cls, map, values):
        if values.get('map_built', False):
            return map
        cls = _get_class_dict(values)

        # cls = get_class(values['kind'], values['type'])
        if map is None:
            map = cls.default_device_map()
        cls.parse_device_map(map)
        return map
    
    
def open_device(cfgfile, path=None, prefix="", key=None, default_type=None, **kwargs):
    """ Open a device from a configuration file 

        
        Args:
            cfgfile: relative path to one of the $CFGPATH or absolute path to the yaml config file 
            key: Key of the created Manager 
            path (str, int, optional): 'a.b.c' will loock to cfg['a']['b']['c'] in the file. If int it will loock to the Nth
                                        element in the file
            prefix (str, optional): additional prefix added to the name or key

        Output:
            device (BaseDevice subclass) :tanciated Device class     
    """
    kwargs.setdefault("kind", KINDS.DEVICE)

    return open_object(cfgfile, path=path, prefix=prefix, key=key, default_type=default_type, **kwargs) 




def parse_device_map(parentClass, map):   
    for name, conf in map.items():
        if isinstance(conf, BaseModel): continue
        
        if isinstance(conf, str):
            conf = io.load_config(conf)
        elif 'cfgfile' in conf:
            tmp = IOConfig.parse_obj(conf)
            conf = io.load_config(tmp.cfgfile)
            conf.update(**tmp.dict(exclude=set(['cfgfile'])))
            
        try:
            prop = getattr(parentClass, name)
        except AttributeError:
            type_ = conf.get('type', None)
            if type_ is None:
                if isinstance(parentClass, BaseDevice):
                    # assume a default type for a device child of a device the parent type 
                    cls = parentClass
                else:
                    cls = parentClass.Device                
            else:
                cls = get_device_class(type_)
            
            if 'node_map' in conf: 
                parse_node_map(cls, conf['node_map'])
            if 'rpc_map' in conf: 
                parse_rpc_map(cls, conf['rpc_map'])
            if 'interface_map' in conf: 
                parse_interface_map(cls, conf['interface_map'])
            if 'device_map' in conf: 
                parse_device_map(cls, conf['device_map'])
            conf['map_built'] = True
            map[name] = cls.Config.parse_obj(conf)
        else:
            # TODO treat the node ?????
            newconf = {**prop._config.dict(), **conf}
            if 'node_map' in newconf: 
                parse_node_map(prop._cls, newconf['node_map'])
            if 'rpc_map' in newconf: 
                parse_rpc_map(prop._cls, newconf['rpc_map'])
            if 'interface_map' in newconf: 
                parse_interface_map(prop._cls, newconf['interface_map'])
            if 'device_map' in newconf: 
                parse_device_map(prop._cls, newconf['device_map'])
            newconf['map_built'] = True
            map[name] = prop._cls.Config.parse_obj(newconf)
                                    
                                                
    for subcls in parentClass.__mro__:
        for name, obj in subcls.__dict__.items():
            if isinstance(obj, _BaseProperty):
                if issubclass(obj._cls, BaseDevice):
                    if not name in map:
                        newconf = obj._config.copy(deep=True).dict()                      
                        if 'node_map' in newconf:
                            parse_node_map(obj._cls, newconf['node_map'] )
                        if 'rpc_map' in newconf:
                            parse_rpc_map(obj._cls, newconf['rpc_map'] )
                        if 'interface_map' in newconf:
                            parse_interface_map(obj._cls, newconf['interface_map'] )
                        if 'device_map' in newconf:
                            parse_device_map(obj._cls, newconf['device_map'] )
                        newconf['map_built'] = True
                        map[name] = obj._cls.Config.parse_obj(newconf) 
    
    
class DeviceProperty(_BaseProperty):    
    fbuild = None    
    
    def builder(self, func):
        """ Decorator for the interface builder """
        self.fbuild = func
        return self
    
    def get_config(self, parent):
        try:
            sconf = parent.config.device_map[self._name]
        except (KeyError, AttributeError):
            return self._config.copy(deep=True)
        else:
            return sconf
    
    def __call__(self, func):
        """ The call is used has fbuild decorator 
        
        this allows to do
        
        ::
            
            class MyManager(BaseManager):
                @MyDevice.prop('motor2')
                def motor2(self, motor):
                    # do somethig
                    
        """
        self.fbuild = func
        return self
    
    def _finalise(self, parent, device):
        # overwrite the fget, fset function to the node if defined         
        if self.fbuild:
            self.fbuild(parent, device)  
                      
class BaseDevice(_BaseObject, _NodeListing, _InterfaceListing, _RpcListing):
    Property = DeviceProperty
    Config = BaseDeviceConfig
    Interface = BaseInterface
    Data = BaseData
    Node = BaseNode
    Rpc = BaseRpc    
    
    _com = None                
    def __init__(self, 
           key: Optional[str] = None, 
           config: Optional[Config] = None,
           com: Optional[Any] = None,             
           **kwargs
        ) -> None:        
        
        super().__init__(key, config=config, **kwargs)  
        if self._localdata is None:
            self._localdata = {}
        self._com = self.new_com(self._config, com)
    
    @classmethod
    def parse_config(cls, config, **kwargs):
        if isinstance(config, dict):
            kwargs = {**config, **kwargs}
            config = None
        if config is  None:
            kwargs['map_built'] = True
            parse_device_map(cls, kwargs.setdefault('device_map', {}))
            parse_interface_map(cls, kwargs.setdefault('interface_map', {}))
            parse_node_map(cls, kwargs.setdefault('node_map', {}))
            parse_rpc_map(cls, kwargs.setdefault('rpc_map', {}))
            
        return super().parse_config(config, **kwargs)
        
    @classmethod
    def new_com(cls, config: Config, com: Optional[Any] = None) -> Any:
        """ Create a new communication object for the device 
            
        Args:
           config: Config object of the Device Class to build a new com 
           com : optional, A parent com object used to build a new com if applicable  
           
        Return:
           com (Any): Any suitable communication object  
        """
        return com 
    
        
    def __getattr__(self, attr):
        try:
            return object.__getattribute__(self, attr)
        except AttributeError:
            try:
                return self._build_interface(attr)
            except ChildError:
                pass
            try:
                return self._build_node(attr)
            except ChildError:
                pass
            try:
                return self._build_rpc(attr)
            except ChildError:
                pass
            try:
                return self._build_device(attr)
            except ChildError:
                pass
            raise AttributeError(attr)
        
    def clear(self):
        """ clear all cashed intermediate objects """
        self._clear_all()  
            
    @property
    def com(self):
        return self._com
        
    def connect(self):
        """ Connect device to client """
        raise NotImplementedError('connect method not implemented') 
    
    def disconnect(self):
        """ Disconnect device from client """
        raise NotImplementedError('disconnect method not implemented')    
    
    def is_connected(self):
        """ True if device connected """
        raise NotImplementedError('is_connected method not implemented') 
    
    def rebuild(self):
        """ rebuild will disconnect the device and create a new com """
        self.disconnect()
        self.clear()
        self._com = self.new_com(self._config)
    
        
class _DeviceListing:
    # provide a standard capability to have node has children with some method
    # in this context self is the parent 
    
    
    def _config_device_keys(self):
        return self.config.device_map.keys()
    
    def _config_device_constructor(self, name):
        try:
            deviceconf = self.config.device_map[name]
        except KeyError:
            raise ChildError(f"Unknown device {name}")
        return get_device_class(deviceconf.type).new    
    
    def _build_device(self, name):
        constructor = self._config_device_constructor(name)
        device = constructor(self, name, config=self.config.device_map[name])
        self.__dict__[name] = device 
        return device    
    
    parse_device_map = classmethod(parse_device_map)
    
    @classmethod
    def default_device_map(cls):
        return {}
    
    @property
    def devices(self) -> List[BaseDevice]:
        """ An iterator on devices """
        return ObjectIterator(self, self._build_device, self._config_device_keys())
        
    def get_device(self, name: str):
        """ get the device child 
        
        Args:
           name (str): device recorded name (attribute of parent)
           
        """
        try:
            #cached node
            device = self.__dict__[name]
        except KeyError:
            try:
                # static device, as property 
                device = object.__getattribute__(self, name)
            except AttributeError:
                device = self._build_device(name)                
            else:
                if not isinstance(device, BaseDevice):
                    raise ValueError(f"Unknown device {name}") 
        else:
            if not isinstance(device, BaseDevice):
                raise ValueError(f"Unknown device {name}")        
        return device
    
    def add_device(self, name, device: Union[BaseDevice, BaseDevice.Config, dict]):
        """ Add a device
        
        This will also change the manager config.device_map 
        
        Args:
            name (str): The name (attribute) of the device
                
                ::
                
                    parent.add_device('motor1', motor1)
                    parent.motor1 is motor1
            
            device (Device, Config , dict): A device object or a config to build the new device
        
        """    
        if isinstance(device, dict):
            type_ = device.get('type',None)
            if type_ is None:
                cls = self.Device
            else:
                cls = get_class(KINDS.DEVICE, type_)
            if name is None:
                raise ValueError("name is missing")                                
            device = cls.new(self, name, **device)
        elif isinstance(device, BaseModel):
            if name is None:
                raise ValueError("name is missing") 
            device = cls.new(self, name, config=device)
            
        
        if name is None:
            _, name = ksplit(device.key)
        
        if name in self.config.device_map and self.config.device_map[name] != device.config:
            raise ValueError("Device already exists inside its parent")
        elif name in self.__dict__:
            raise ValueError(f"attribute {name} is already taken")
        
        self.config.device_map[name] = device.config
        setattr(self, name, device)  
      
    def remove_device(self, name):
        """ Remove a device 
        
        This will remove also the configuration inside the config.device_map 
        An ValueError is raised if the device is not in the parent or if it has been defined 
        statisticaly on the class.
        
        Args:
            name (str) : the reference (the attribute) of the device
        """
        
        try:
            getattr(self.__class__, name)
        except AttributeError:
            if not name in self.config.device_map:
                raise ValueError("Unknown device {name!r}")
            
            del self.config.device_map[name]
            try:
                del self.__dict__[name]
            except KeyError:
                pass    
        else:
            raise ValueError("device {name!r} was defined inside the class")    
        
                
    
# Add the _DeviceListing  capability to the BaseDevice    
@record_class
class BaseDevice(BaseDevice, _DeviceListing):
    pass        
    
    
    



from pydevmgr_core import *

if False:
    from matplotlib.pylab import plt
    use_plot = True
else:
    use_plot = False

def test_main():
    
    try:
        import numpy
    except ModuleNotFoundError:
        print( "Warning numpy related node cannot be tested : numpy not installed" )
        return 
    
    s = StaticNode(value=16)
    v = NoiseAdderNode(node=s, scale=4)

    st = StatisticNode(node=v, mean=16)

    h = HistogramNode(node=v, bins=(0,32,33))
    for i in range(10000):
        _st, _h = download( [st, h] )
        
    print(_st)
    if use_plot:
        plt.bar( (h._bins[:-1]+h._bins[1:])/2,_h)
        plt.axvline(_st.mean, color="red")
        plt.show()

if __name__ == "__main__":
    test_main()



import pkg_resources
import os

class PkgResource:
    def __init__(self, pkg_name, resource_dir="resources"):
        self._pkg_name = pkg_name
        self._resource_dir = resource_dir

    @property
    def directory(self):
        return pkg_resources.resource_filename(
                self._pkg_name, 
                self._resource_dir
                )
    def exists(self, resource_name):
        rpath = os.path.join(self._resource_dir, resource_name) 
        return pkg_resources.resource_exists(self._pkg_name, rpath)
    
    def find_resource(self, resource_name):
        if not self.exists(resource_name):
            raise ValueError(f"package resource {resource_name} not found ")
        rpath = os.path.join(self._resource_dir, resource_name) 
        return pkg_resources.resource_filename(self._pkg_name, rpath)


if __name__ == "__main__":
    pkg = PkgResource("pydevmgr_core")
    print('PACKAGE DIRECTORY', pkg.directory)
    pkg.exists('test_node.yml')
    print('test_node file is their:', pkg.find_resource('test_node.yml'))        



from pydevmgr_core import BaseDevice, StaticNode, BaseInterface, to_nodealias_class, to_rpc_class
from pydevmgr_core import NodeAlias, NodeAlias1, record_class, get_class, BaseRpc, FormulaNode

import yaml


@record_class
class Ctrl(BaseInterface):
    class Config(BaseInterface.Config):
        type  = "Control"
        description = "This is something"
    go = StaticNode.prop('go', value=True)


@record_class(overwrite=True)
class Printer(BaseRpc):
    class Config(BaseRpc.Config):
        type  = "Printer"
    def fcall(self, *args):
        print(self.key, 'called with args', args)

@record_class(overwrite=True)
@to_rpc_class
def Printer2(key, *args):
    print(key, 'called with args', args)


@record_class(overwrite=True)
class Scaler(NodeAlias1):
    class Config(NodeAlias1.Config):
        scale: float  = 100.0
        type: str =  "Scaler"
        dummy: int = 1
        
    def fget(self, value):
        return self.config.scale*value

@record_class(overwrite=True)
@to_nodealias_class
def Scaler2(value, scale:float = 100.0, dummy: int = 2):
    return value * scale
        
@record_class(overwrite=True)            
class MyDevice(BaseDevice):
    class Config(BaseDevice.Config):
        type: str = "MyDevice"    
    
    purestatic    = StaticNode.prop(value=9) # try without a name 
    changedstatic = StaticNode.prop('chs', value=9) # with a name 
    @NodeAlias1.prop('plus', 'purestatic')
    def plus(self, ps):
        return ps+10
    
    hundred = Scaler.prop('hundred', 'purestatic')
    fifty = get_class("Node", "Scaler2").prop('fifty', 'purestatic', scale=80, dummy=2)
    
    cfg = BaseInterface.prop('cfg', node_map={'alpha':{'type':'Static', "value":9}})
    ctrl = Ctrl.prop('ctrl')
    
    
cfg_txt = """---
node_map:
    temperature_volt: 
       kind: Node
       type: Static
       value: 1.0
    temperature:
       type: Formula
       formula: 24.0 * t - 2.0
       nodes: ['temperature_volt']
       varnames: [t]
    humidity: 
       kind: Node
       type: Static
       value: 54.0
    chs:
       kind: Node
       type: Static
       value: 19
    hundred:
        type: Scaler
    fifty:
        type: Scaler2
        scale: 50
    fifty2:
        type: Scaler
        scale: 50
        node: temperature
       
rpc_map:
    echo:
        type: Printer    
    echo2:
        type: Printer2    
interface_map:
    stat:
        kind: Interface
        node_map:
            status:
                kind: Node
                type: Static
                value: 10
    ctrl2:
        type: Control
        description: This is something else    
"""

def test_main():  
    config = MyDevice.Config.parse_obj(yaml.load(cfg_txt, yaml.CLoader))
    device = MyDevice('test', config)
    
    
    #device.temperature
    device.get_node('temperature')
    assert device.temperature.get() == 22.0
    assert device.changedstatic.get() == 19
    #assert list(device.nodes) == [device.temperature, device.humidity, device.changedstatic, device.purestatic]    
    device.nodes
    assert device.humidity.get() == 54.0
    assert device.purestatic.get() == 9
    
    assert device.changedstatic.get() == 19
    assert device.plus.get() == 19 
    assert device.stat.status is device.stat.status
    
    stat = device.stat
    assert stat is device.get_interface('stat')
    
    #stat = device.get_interface('stat')
    #print(stat.config)
    
    assert stat.get_node('status').get() == 10
    assert stat.status.get() == 10
    assert config.interface_map['stat'].node_map['status'] is device.stat.status.config
    assert device.cfg.alpha.get() == 9
    assert device.ctrl.go.get() == True
    assert device.ctrl2.go.get() == True
    assert device.ctrl.config.description == "This is something"
    assert device.ctrl2.config.description == "This is something else"
    
    device.echo.call("un", 2, "trois")
    device.echo2.call("un", 2, "trois")
    #assert device.interfaces == []
    device.hundred.get() == device.purestatic.get()*100
    device.fifty.get() == device.purestatic.get()*50
    device.fifty2.get() == device.temperature.get()*50
    
    assert device.hundred.config is device.config.node_map['hundred']
    
    assert device.purestatic.config is not MyDevice.purestatic._config
    
    assert device.fifty.config.dummy == 2
    assert device.config.node_map['fifty'] is device.fifty.config
    print('test_device', 'OK')

if __name__ == "__main__":
    test_main()



from typing import Any,  Tuple, Optional, List, Dict, Union, Type
from pydantic import BaseModel, Extra
from ._class_recorder import get_class
from ._core_model_var import StaticVar
from enum import Enum 
import yaml
from ..io import ioconfig, load_yaml, load_config


class KINDS(str, Enum):
    PARSER = "Parser"
    NODE = "Node"
    RPC = "Rpc"
    DEVICE = "Device"
    INTERFACE = "Interface"
    MANAGER = "Manager"
    


class IOConfig(BaseModel):
    """ model for cfgfile definition """    
    cfgfile: str = ""
    class Config:
        extra = Extra.allow
    
class BaseConfig(BaseModel):
    kind: KINDS = ""
    type: str = ""
    version: str = "" # version of the configuration file
    true_type: Optional[Type] = None
    
    class Config: # this is the Config of BaseModel
        extra = Extra.forbid
        validate_assignment = True    
        use_enum_values = True 
    
    @classmethod
    def from_cfgfile(cls, cfgfile, path: str = ''):
        """ Create the config object from a yaml config file 
        
        Args:
            cfgfile (str): Configuration file relative to one of $CFGPATH or absolute path 
            path (str, optional): path where to find the root configuration. if "" the config file 
                 contains the root. For instance 'a.b.c' will look at configuration in cfg['a']['b']['c']
        """
        config = load_config(cfgfile)
        for p in (s for s in  path.split('.') if s):
            config = config[p]
        return cls.from_cfgdict(config)                    
    
    @classmethod 
    def from_cfgdict(cls, config_dict):
        """ Create the config object from python dictionary """
        return cls.parse_obj(config_dict)
         
    
    def cfgdict(self, exclude: set =set()):
        """ Write the config to a dictionary as it is red from from_cfgdict """
        exclude.add('true_type')
        d = self.dict(exclude_unset=True, exclude=exclude)
        return d

def _path_walk(d, path):
    if isinstance(path, int):
        # get the first key
        
        if hasattr(d, "keys"):
            k = list(d.keys())[path]
            return d[k]
        else:
            return d[path]
   
    if path:
        for p in (s for s in  path.split('.') if s):
            d = d[p]
    return d

def _get_class_dict( d, default_type: Optional[Union[str, Type]] = None):

    try:
        kind = d['kind']
    except KeyError:
        raise ValueError('"kind" attribute missing')
    
    tt = d.get('true_type', None)
    if tt: 
        return tt
    
    try:
        st = d['type']
    except KeyError:
        raise ValueError('"type" attribute missing')    

    try:
        return get_class(kind, st)
    except ValueError as e:
        if default_type:
            if isinstance(default_type, type):
                return default_type
            return get_class(kind, default_type)
        raise e


def _get_class_config( c, default_type: Optional[Union[str, Type]] = None):
    if c.true_type:
        return c.true_type
    try:
        return get_class(c.kind, c.type)
    except ValueError as e:
        if default_type:
            if isinstance(default_type, type):
                return default_type
            return get_class(c.kind, default_type)
        raise e
  




def open_class(
        cfgfile: str, 
        path: Optional[Union[str, int]] = None, 
        default_type: Optional[str] = None,
        **kwargs
        ):
    """ open a pydevmgr class and configuration object from a config file 

    Args:
        cfgfile: relative to on of the $CFPATH or absolute path to yaml config file 
        kind (optional, str): object kind as enumerated in KINDS ('Manager', 'Device', 'Interface', 'Node', 'Rpc')
            if None look inside the configuration file and raise error if not defined. 
        
        path (optional, str, int): an optional path to find the configuration inside the config file 
             'a.b.c' will go to cfg['a']['b']['c']
             If an integer N will get the Nth element of the cfgfile 

        default_type (optional, str): A default type if no type is defined in the configuration file
            If default_type is None and no type is found an error is raised 

    Returns:
        ObjClass :  An pydevmgr Object class (Manager, Device, Node, Rpc, Interface)
        config : An instance of the config (BaseModel object) 
    """
    allconf = load_config(cfgfile)
        
    allconf = _path_walk(allconf, path)
    
    allconf.update( kwargs )
    
    try:
        kind = allconf['kind']
    except KeyError:
        raise ValueError("Configuration has no 'kind' defined")

    tpe = allconf.get('type', default_type)
    if not tpe:
        raise ValueError(f"Cannot resolve {kind} type")
    Object = get_class(kind, tpe)
    allconf['true_type'] = Object
    # config = Object.parse_config(allconf) 
    config = Object.Config.from_cfgdict(allconf)
    return Object, config

def open_object(
        cfgfile,
        key: Optional[str]= None, 
        path: Optional[Union[str, int]] = None, 
        prefix: str = '', 
        default_type: Optional[str] = None, 
        **kwargs
    ):
    """ open an object from a configuration file 
    Args:
        cfgfile: relative to on of the $CFPATH or absolute path to yaml config file 
        kind (optional, str): object kind as enumerated in KINDS ('Manager', 'Device', 'Interface', 'Node', 'Rpc')
            if None look inside the configuration file and raise error if not defined. 
        
        path (optional, str): an optional path to find the configuration inside the config file 
             'a.b.c' will go to cfg['a']['b']['c']
        default_type (optional, str): A default type if no type is defined in the configuration file
            If default_type is None and no type is found an error is raised

    Returns:
        obj : instanciatedpydevmgr object (Manager, Device, Node, Rpc, Interface)

    """
    Object, config = open_class(cfgfile, path=path, default_type=default_type, **kwargs)
    if key is None and isinstance(path, str): 
        _, name = ksplit(path)           
        key = kjoin(prefix, name)
    return Object(key, config=config)
        
class BaseData(BaseModel):
    # place holder for Data class 
    key: StaticVar[str] = ""

class ChildError(ValueError):
    """ Error raise when a parent try to reach a child object (interface, node, rpc, ...) and cannot find it"""
    pass

def load_yaml_config(yaml_payload: str, path: Optional[Union[str, tuple]] = None) -> Tuple[Type,BaseConfig]:
    """ Load a yaml configuration and pare it in the right configuration object 
    
    The Config class used is localised thanks to the `kind` and `type` string argument inside the yaml
    If the class is not recognised an Exception is raised.
    
    Args:
        yaml_payload (str):  The yaml string payload 
        path (optional, str, tuple):  A string or tuple representing a path through the wanted object 
    
    Returns:
        cls (Type):  The object Class 
        config (BaseModel):  The parsed configuration 
    """
    payload = yaml.load(yaml_payload, Loader=ioconfig.YamlLoader)
    
    payload = _path_walk(payload, path)
   
    return load_dict_config(payload)
    
def load_dict_config(payload: dict)-> Tuple[Type,BaseConfig]:
    try:
        kind = payload['kind']
    except KeyError:
        raise ValueError('"kind" attribute missing')
    try:
        type = payload['type']
    except KeyError:
        raise ValueError('"type" attribute missing')    
    
    cls = _get_class_dict(payload)
    return cls, cls.parse_config(payload)

def build_yaml(yaml_payload, key: Optional[str]=None, *args, **kwargs):
    """ Build and return an object from its yaml configuration payload """
    cls, config = load_yaml_config(yaml_payload)
    return cls(key, *args, config=config, **kwargs)

def load_and_build( cfgfile: str, key: Optional[str] = None, *args, **kwargs):
    """ Load a config file and build the conrespoding object defined by kind and type config parameters """
    payload = load_config(cfgfile)
    
    cls, config = load_dict_config(payload)
    return cls(key, *args, config=config, **kwargs)
    
def kjoin(*args) -> str:
    """ join key elements """
    return ".".join(a for a in args if a)

def ksplit(key: str) -> Tuple[str,str]:
    """ ksplit(key) ->  prefix, name
    
    >>> ksplit('a.b.c')
    ('a.b', 'c')
    """
    s, _, p = key[::-1].partition(".")
    return p[::-1], s[::-1]

def reconfig(ConfigClass: Type, config: BaseConfig, kwargs: dict) -> BaseConfig:    
    if config is None:
        return ConfigClass.parse_obj(kwargs)
    if isinstance(config, dict):
        return ConfigClass.parse_obj(dict(config, **kwargs))    
    return config


_key_counter = {}
def new_key(config):
    k = config.type+config.kind
    c = _key_counter.setdefault(k,0)
    c += 1
    _key_counter[k] = c
    return f'{k}{c:03d}'


class _BaseProperty:    
    """ A Property is basically calling a constructor with dynamical and static arguments 
    
    The Property has the followind signature : 
        Property(constructor, name, *args, config=None, **kwargs)
        
    The constructor is called with the following signature
        constructor( parent , name, *args, config=config, **kwargs)
    
    So it must have at least twho positional arguments a config keyword argument and 
    optional keyword arguments  
    
    """    
    def __init__(self, cls, constructor, name, *args, config=None,  **kwargs):
        
        self._cls = cls 
        self._constructor = constructor
        self._name = name         
        
        self._config = config        
        self._args = args
        self._kwargs = kwargs
        
    @property
    def congig(self):
        return self._config
            
    def _finalise(self, parent, obj):
        pass
    
    def __get__(self, parent, clp=None):
        if parent is None:
            return self 
        # try to retrieve the node directly from the parent __dict__ where it should 
        # be cached. If no boj cached create one and save it/ 
        # if _name is the same than the attribute name in class, this should be called only ones                
        try:
            obj = parent.__dict__[self]
        except KeyError:
            name, obj = self.new(parent)     
            # store in parent with the name 
            parent.__dict__[self] = obj        
        return obj
                
    def get_config(self, parent):
        """ return configuration from parent object """
        # this has to be implemented for each kinds 
        return self._config.copy(deep=True)
                
    def new(self, parent):     
        config = self.get_config(parent)
        if self._name is None:
            name = new_key(config)
            
            #name = config.kind+str(id(config))
        else:
            name = self._name                            
        obj = self._constructor(parent, name, *self._args, config=config, **self._kwargs)            
        self._finalise(parent, obj)
        return name, obj 
        

class _BaseObject:
    __all_cashed__ = False
    Config = BaseConfig
    Property = _BaseProperty    
    _config = None
            
    def __init__(self, 
          key: Optional[str] = None,  
          config: Optional[Config] = None, *,          
          localdata: Optional[dict] = None, 
          **kwargs 
        ) -> None:
        self._config = self.parse_config(config, **kwargs)                            
        if key is None: 
            key = new_key(self._config)
        
        self._key = key     
        self._localdata = localdata
        
    def __repr__(self):
        return "<{} key={!r}>".format(self.__class__.__name__, self._key)
    
    
    @classmethod
    def parse_config(cls, __config__=None, **kwargs):
        if __config__ is None:
            kwargs['true_type'] = cls
            return cls.Config(**kwargs)
        if isinstance(__config__ , cls.Config):
            return __config__
        if isinstance(__config__, dict):
            d = {**__config__, **kwargs} 
            d['true_type'] = cls

            # if hasattr(cls.Config, "from_cfgdict"):
            #     return cls.Config.from_cfgdict(d)
            return cls.Config( **d )
        raise ValueError(f"got an unexpected object for config : {type(__config__)}")
            
    @classmethod
    def new_args(cls, parent, config: Config) -> dict:
        """ build a dictionary of dynamical variables inerited from a parent """
        return dict( localdata = getattr(parent, "localdata", None) )
            
    @classmethod
    def new(cls, parent, name, config=None, **kwargs):
        """ a base constructor for BaseNode 
        
        parent must have the .key attribute 
        """        
        # here shall be implemented something to deal with config, it might be that config it comming
        # from the parent.config like node_map, etc ...
        config = cls.parse_config(config, **kwargs)
        if name is None:
            name = new_key(config)                                
        return cls(kjoin(parent.key, name), config=config, **cls.new_args(parent, config))
    
    @classmethod
    def prop(cls, name: Optional[str] = None, **kwargs):
        """ Return an object  property  to be defined in a class 
        
        Exemple:
           
           ::
           
                def MyDevice(BaseDevice):
                   ref_temperature = StaticNode.prop(value=22.0
                
                MyDevice().ref_temperature.get()   
        """
        config = cls.Config.parse_obj(kwargs)                
        return cls.Property(cls, cls.new, name, config=config)
    
    @classmethod
    def from_cfgfile(cls, 
            cfgfile: str, 
            key: Optional[str]= None, 
            path: Optional[Union[str,int]] = None, 
            prefix: str = '', 
            **kwargs
        ) -> '_BaseObject':
        """ Create the object from a configuration file 
        
        Args:
            cfgfile (str): Path to the config file, shall be relative to one of the $CFGPATH or absolute
                     path
            key (str, Optional): key of the device, if not given this is built from the path suffix and
                    the optional prefix
           
            path (None, str, False)"" the hierarchique path where to find the config data inside the file 
                    for instance 'a.b.c' will loock at cfg['a']['b']['c'] from the loaded config file 
                    If "" the config file define the device configuration from its root 
                    If None the first item of the config file is taken 
            prefix (str, optional): add a prefix to the path name to build the device key. 
                    It is used only if key is None otherwhise ignored
        """
        
        allconf = load_config(cfgfile)
        allconf = _path_walk(allconf, path)
        
        allconf.update(**kwargs)
               
        if key is None and isinstance(path, str): 
            _, name = ksplit(path)           
            key = kjoin(prefix, name)
        
        return cls.from_cfgdict(allconf, key)
    
    @classmethod
    def from_cfgdict(cls, cfgdict, key: Optional[str] = None, *args, **kwargs):
        """ Open  """
        tpe = cls.Config.__fields__['type'].default 
        #config = cls.parse_config(cfgdict, type=tpe)
        config = cls.Config.from_cfgdict(cfgdict)               
        
        return cls(key, *args, config=config, **kwargs)
        
    @property
    def key(self):
        """ key """
        return self._key
    
    @property
    def config(self):
        """  config """
        return self._config
    
    @property
    def localdata(self):
        """ localdata dictionary """
        return self._localdata
        
    @property
    def name(self):
        return ksplit(self._key)[1]    
    
    @classmethod
    def _builtin_objects(cls, SubCls=None):
        SubCls = SubCls or _BaseObject
        d = {}
        for sub in cls.__mro__:
            for k,v in sub.__dict__.items():
                if isinstance(v, (_BaseProperty)):
                    if issubclass(v._cls, SubCls):
                        d[k] = v._cls
        return d
                    
    def _cash_all(self):
        """ cash all child _BaseProperty to real Objects """
        if not self.__all_cashed__:
            for sub in self.__class__.__mro__:
                for k,v in sub.__dict__.items():
                    if isinstance(v, (_BaseProperty)):
                        getattr(self, k)
            self.__all_cashed__ = True
    
    def _clear_all(self):
        """ clear all cashed intermediate objects """
        for k,v in list(self.__dict__.items()):
            if isinstance(v, (_BaseObject)):
                self.__dict__.pop(k)
        self.__all_cashed__ = False



class ObjectIterator:
    _iterator = None
    def __init__(self, obj, constructor, names):
        self._obj = obj
        self._constructor = constructor
        self._names = names 
    
    def __iter__(self):
        self._iterator = iter(self._names)
        return self
    
    def __next__(self):
        name = next(self._iterator)
        try:
            return object.__getattribute__(self._obj, name)
        except AttributeError:
            return self._constructor(name)
        
    def __getitem__(self, name):
        
        try:
            return object.__getattribute__(self._obj, name)
        except AttributeError:
            return self._constructor(name)
    
    def __call__(self) -> List:
        return list(self)
    
    def names(self) -> List:
        return list(self._names)
       
       
class _Old_ObjectIterator_To_delete:
    _iterator = None
    def __init__(self, map: dict, classes: Union[Type,Tuple[Type]]):
        self._map = map
        self._classes = classes 
    
    def __iter__(self):
        self._iterator = iter(self._map.items())
        return self
    
    def __next__(self):
        k,v = next(self._iterator)
        if isinstance(v, self._classes):
            return v
        return self.__next__()
    
    def __getitem__(self, item):
        v = self._map[item]
        if isinstance(v, self._classes):
            return v 
        raise KeyError(item)
    
    def __call__(self) -> List:
        return list(self)
    
    def names(self) -> List:
        return [k for k,v in self._map.items() if isinstance(v, self._classes)]
       
def _collect_config_properties(cls, prop_kind):
    for sub in cls.__mro__:
        for k,v in sub.__dict__.items():
            if isinstance(v, (_BaseProperty)):
                if v._config.kind is prop_kind:
                    yield k, v._config





                       
        



from pydevmgr_core import BaseDevice, BaseNode, record_class
from pydantic import BaseModel 
from typing import Dict
import yaml


cfg_txt = """---
node_map:
    temperature:        
       type: Static
       value: 22.0
    humidity:    
       type: Static
       value: 54.0
interface_map:
    cfg:
        node_map:
            alpha:
                type: Static
                value: 8.4 
"""

cfg_txt2 = """---
node_map:
    temperature:  
       type: Static
       value: 22.0
    humidity:
        type: Static       
        value: 54.0
"""


cfg_txt_er1 = """---
temperature: 22
"""


def test_main():
    config = BaseDevice.Config.parse_obj(yaml.load(cfg_txt, yaml.CLoader))
    assert config.node_map['temperature'].value == 22.0    
    assert config.interface_map['cfg'].node_map['alpha'].value == 8.4 
    
    config = BaseDevice.Config.parse_obj(yaml.load(cfg_txt2, yaml.CLoader))
    assert config.node_map['temperature'].value == 22.0
    print('test_device_config', 'OK')

if __name__ == "__main__":
    test_main()





from pydantic import create_model, Extra, BaseModel, validator
from typing import Dict, Any, Optional, Union, Callable, Iterable, List, Dict, Tuple
from pydevmgr_core._core_objects._core_parser import create_parser_class, BaseParser, _BuiltParser, Clipped

class AnyParserConfig(BaseModel):
    type: Union[List[Union[str, Callable]], str, Callable] = ""
    class Config:
        extra = Extra.allow
        
class BaseNodeConfig(BaseModel):
    """ Config for a Node """
    kind: str = "Node"
    type: str = ""
    #parser: Optional[Union[str, Callable, Iterable]] = None
    parser: Optional[Union[AnyParserConfig, List[Union[str, Callable]], str, Callable]] = None
    #parser_config: dict = {}
    description: str = ""
    unit: str = ""
    
    class Config:
        validate_assignment = True
        


def parser(parsers, config=None, **kwargs):
    if isinstance(parsers, (BaseParser, _BuiltParser)) and not config and not kwargs:
        return parsers
    
    if isinstance(parsers, AnyParserConfig):
        Parser = create_parser_class(parsers.type)
        kwargs = {**kwargs, **parsers.dict(exclude=set(["type"]))}
    elif isinstance(parsers, dict):
        type_ = parsers.get('type', None)
        if not type_ :
            raise ValueError('parser type cannot be None')
        Parser = create_parser_class(type_)
        kwargs = {**parsers, **kwargs}    
    else:    
        Parser = create_parser_class(parsers)
    return Parser(config=config, **kwargs)        


def test_main():   
    assert parser(BaseNodeConfig(parser=float).parser)('4') == 4.0
    assert parser(BaseNodeConfig(parser={'type':'Float'}).parser)('4')
    assert parser(BaseNodeConfig(parser={'type':['Float','Clipped'], 'max':10}).parser)('12') == 10.0
    assert parser(BaseNodeConfig(parser='Float').parser)('4') == 4.0    
    assert parser(BaseNodeConfig(parser=Clipped(max=10)).parser)(12.0) == 10.0
    
    c = BaseNodeConfig(parser=Clipped(max=10))
    p = parser(c.parser)
    c.parser = p.config
    c.__dict__['parser'] = p.config
    assert c.parser is p.config    
    c.parser.max = 12
    assert p(13) == 12.0


if __name__ == "__main__":
    test_main()
    



from pydevmgr_core import parser, create_parser_class, Rounded, Clipped,  Stripped, ToString
from pydevmgr_core import BaseNode, BaseRpc

def test_main():
    fr = parser((float, Rounded, Clipped), ndigits=2, max=3.0)
    fr2 = parser((float, Rounded, Clipped), dict(ndigits=2, max=3.0))
    assert fr.__class__ is fr2.__class__
    assert fr(2.3456) == 2.35
    assert fr(3.3456) == 3.0
    fr.config.max = 4
    assert fr(3.3456) == 3.35

    assert fr2.config.ndigits == 2

    FR = create_parser_class((float, Rounded, Clipped))
    fr = parser( (FR,  ToString), ndigits=2, max=3.0, format="%.2f")
    assert fr(3.3456) == '3.00'

    fr = parser( (float, ))        
    fr = parser( (Rounded, ), ndigits=0)
    assert fr(1.234)==1.0   


    fr = parser( Rounded, ndigits=0)
    assert fr(1.234)==1.0    
    fr = parser( 'Rounded', ndigits=0)
    assert fr(1.234)==1.0  
    fr = Rounded(ndigits=0)
    assert fr(1.234)==1.0  
      
    fr = parser( ('Rounded','Clipped'), ndigits=0, min=0)
    assert fr(1.234) ==1.0    
    assert fr(-1.234) == 0.0    

    fr = parser((str, Stripped))
    assert fr(' un ') == 'un'
    assert fr(1.3) == '1.3'


    n = BaseNode(parser="Clipped")
    assert n._parser.config is n.config.parser
    n.config.parser.max = 12 
    assert  n._parser.config.max  == 12

    r = BaseRpc(args_parser=["Clipped"], kwargs_parser={'toto':[float,"Clipped"] })
    assert r._args_parser[0].config is r.config.args_parser[0]
    r.config.args_parser[0].max = 10
    assert r._kwargs_parser['toto'].config is r.config.kwargs_parser['toto']

if __name__ == "__main__":
    test_main()



from pydevmgr_core import * 
from pydantic import Field 

class D(BaseDevice):
    #r = NoiseNode.prop('r', distribution="random", scale=0.8, mean=0.4)
    r = StaticNode('r', value=0.4)
    @NodeAlias1.prop('r2', node='r')
    def r2(self, value):
        return round(value, 2)
        
    posname = PosNameNode.prop('posname', node='r', poses={'A':0.2, 'B':0.4, 'C':0.6}, tol=0.2)
        
class Data(BaseDevice.Data):
    q : NodeVar[list] = Field([], node=DequeNode1.prop('q', node='r2'))
    rbis : NodeVar[float] = Field(0.0, node='r2')
    r : NodeVar[float] = 0.0
    r2 : NodeVar[float] = 0.0
    posname : NodeVar[list] = Field([], node=DequeNode1.prop('posname_list', node='posname'))


def test_main():
    
    data = Data()
    dev = D('d')
    
    dl = DataLink(dev, data)
    
    dl.download()
    dl.download()
    dl.download()
    
    dl.reset()
    
    dl.download()
    dl.download()
    dl.download()
    
    print(data)

if __name__ == "__main__":
    test_main()



from ._core_base import (_BaseObject, _BaseProperty,  KINDS,
                          ObjectIterator, ChildError, IOConfig, ksplit, BaseData, 
                        _get_class_dict
                        )
                         
from ._class_recorder import get_interface_class, get_class, record_class

from ._core_com import BaseCom
from ._core_node import BaseNode, _NodeListing, parse_node_map
from ._core_rpc import BaseRpc, _RpcListing, parse_rpc_map
from .. import io
from typing import Optional, Iterable, Union, List, Dict, Callable
from pydantic import BaseModel, validator, Extra
#  ___ _   _ _____ _____ ____  _____ _    ____ _____ 
# |_ _| \ | |_   _| ____|  _ \|  ___/ \  / ___| ____|
#  | ||  \| | | | |  _| | |_) | |_ / _ \| |   |  _|  
#  | || |\  | | | | |___|  _ <|  _/ ___ \ |___| |___ 
# |___|_| \_| |_| |_____|_| \_\_|/_/   \_\____|_____|
# 


class BaseInterfaceConfig(_BaseObject.Config):
    """ Config for a Interface """
    kind: KINDS = KINDS.INTERFACE.value
    type: str = "Base"    
    map_built: bool = False
    node_map : Dict[str, Union[BaseNode.Config, dict]] = {}
    rpc_map : Dict[str, Union[BaseRpc.Config, dict]] = {} 
    
    @validator('rpc_map', pre=True, always=True)
    def _validate_rpc_map(cls, map, values):
        if values.get('map_built', False):
            return map
        cls = _get_class_dict(values)

        # cls = get_class(values['kind'], values['type'])
        if map is None:
            map = cls.default_rpc_map() 
        cls.parse_rpc_map(map)
        return map
    
    @validator('node_map', pre=True, always=True)
    def _validate_node_map(cls, map, values):
        if values.get('map_built', False):
            return map
        cls = _get_class_dict(values)

        # cls = get_class(values['kind'], values['type']) 
        if map is None:
            map = cls.default_node_map()       
        cls.parse_node_map(map)
        return map
    
    
    def cfgdict(self, exclude=set()):                
        all_exclude = {*{"map_built", "node_map", "rpc_map"}, *exclude}
        d = super().cfgdict(exclude=all_exclude)
        for map  in ('node_map', 'rpc_map'):
            if map not in exclude:
                d[map] = {k:n.cfgdict() for k,n in getattr(self, map).items()}
        
        return d
        
    
    
    class Config: # this is the Config of BaseModel
        extra = Extra.forbid        


def parse_interface_map(parentClass, map):        
    for name, conf in map.items():
        if  isinstance(conf, BaseModel): continue
        
        if isinstance(conf, str):
            conf = io.load_config(conf)
        elif 'cfgfile' in conf:
            tmp = IOConfig.parse_obj(conf)
            conf = io.load_config(tmp.cfgfile)
            conf.update(**tmp.dict(exclude=set(['cfgfile'])))
                    
        try:
            prop = getattr(parentClass, name)
        except AttributeError:
            type_ = conf.get('type', None)
            if type_ is None:
                cls = parentClass.Interface
            else:
                cls = get_interface_class(type_)
            
            if 'node_map' in conf: 
                parse_node_map(cls, conf['node_map'])
            if 'rpc_map' in conf: 
                parse_rpc_map(cls, conf['rpc_map'])
            conf['map_built'] = True
            map[name] = cls.Config.parse_obj(conf)
        else:
            # TODO treat the node ?????
            newconf = {**prop._config.dict(), **conf}
            if 'node_map' in newconf: 
                parse_node_map(prop._cls, newconf['node_map'])
            if 'rpc_map' in newconf: 
                parse_rpc_map(prop._cls, newconf['rpc_map'])
            newconf['map_built'] = True
            map[name] = prop._cls.Config.parse_obj(newconf)
                                    
                                              
    for subcls in parentClass.__mro__:
        for name, obj in subcls.__dict__.items():
            if isinstance(obj, _BaseProperty):
                if issubclass(obj._cls, BaseInterface):
                    if not name in map:
                        map[name] = obj._config.copy(deep=True).dict()                        
                        if 'node_map' in map[name]:
                            parse_node_map(obj._cls, map[name]['node_map'] )
                        if 'rpc_map' in map[name]:
                            parse_rpc_map(obj._cls, map[name]['rpc_map'] )
                        map[name]['map_built'] = True    
                        map[name] = obj._cls.Config.parse_obj(map[name]) 
    
    
    
def annotation2node(cls, PropertyClass, arg=None):
    if isinstance(PropertyClass, str):
        PropertyClass = getattr(cls, PropertyClass).prop
    
    if arg is None:
        def mk_kwargs(p):
            return p
        def ckeck(p):
            return isinstance (p, dict)
                                
    elif isinstance(arg, str):
        def mk_kwargs(p):
            if isinstance (p, dict):
                return p
            return {arg:p}
        def check(p):
            return True
        
    else:
        def mk_kwargs(plist):
            if isinstance (plist, dict):
                return plist
            
            return {k:p for k,p in zip(arg, plist)}
        def check(p):
            return True
        
    for a,p in cls.__annotations__.items():        
        if check(p):
            kwargs = mk_kwargs(p)
            setattr(cls, a, PropertyClass(a, **kwargs))    

def buildproperty(PropertyClass, arg):
    def builder(cls):
        annotation2node(cls, PropertyClass, arg)
        return cls
    return builder 
    
class InterfaceProperty(_BaseProperty):    
    fbuild = None
    def builder(self, func):
        """ Decorator for the interface builder """
        self.fbuild = func
        return self

    def get_config(self, parent):
        try:
            sconf = parent.config.interface_map[self._name]
        except (KeyError, AttributeError):
            return self._config.copy(deep=True)
        else:
            return sconf
                
    def __call__(self, func):
        """ The call is used has fget decorator """
        self.fbuild = func
        return self
    
    def _finalise(self, parent, interface):
        # overwrite the fget, fset function to the node if defined         
        if self.fbuild:
            self.fbuild(parent, interface)            
            
@record_class # we can record this type because it should work as standalone        
class BaseInterface(_BaseObject,  _NodeListing, _RpcListing):
    """ BaseInterface is holding a key, and is in charge of building nodes """    
    
    _subclasses_loockup = {} # for the recorder 
    
    Config = BaseInterfaceConfig
    Property = InterfaceProperty
    Data = BaseData
    Node = BaseNode
    Rpc = BaseRpc   
    
    def __init__(self, 
           key: Optional[str] = None, 
           config: Optional[Config] = None,            
           **kwargs
        ) -> None:        
        
        super().__init__(key, config=config, **kwargs)  
        if self._localdata is None:
            self._localdata = {}
    
    def __getattr__(self, attr):
        try:
            return object.__getattribute__(self, attr)
        except AttributeError:           
            try:
                return self._build_node(attr)
            except ChildError:
                pass
            try:
                return self._build_rpc(attr)
            except ChildError:
                pass
            raise AttributeError(attr)            
                
    def clear(self):
        """ clear all cashed intermediate objects """
        self._clear_all()        

                    
        
class _InterfaceListing:
    # provide a standard capability to have node has children with some method
    # in this context self is the parent 
    
    def _config_interface_keys(self):
        return self.config.interface_map.keys()
    
    def _config_interface_constructor(self, name):
        try:
            interfaceconf = self.config.interface_map[name]
        except KeyError:
            raise ChildError(f"Unknown interface {name}")
        return get_interface_class(interfaceconf.type).new    
        
    def _build_interface(self, name):
        constructor = self._config_interface_constructor(name)
        interface = constructor(self, name, config=self.config.interface_map[name])
        self.__dict__[name] = interface 
        return interface
    
    parse_interface_map = classmethod(parse_interface_map)        
    
    @classmethod
    def default_interface_map(cls):
        return {}
    
    
    @property
    def interfaces(self) -> List[BaseInterface]:
        return ObjectIterator(self, self._build_interface, self._config_interface_keys())
        
    def get_interface(self, name: str) -> BaseInterface:
        try:
            #cached node
            interface = self.__dict__[name]
        except KeyError:
            try:
                # static interface, as property 
                interface = object.__getattribute__(self, name)
            except AttributeError:
                interface = self._build_interface(name)
            else:
                if not isinstance(interface, BaseInterface):
                    raise ValueError(f"Unknown interface {name}")
        else:
            if not isinstance(interface, BaseInterface):
                raise ValueError(f"Unknown interface {name}")           
        return interface
        
    def add_interface(self, name, interface: Union[BaseInterface, BaseInterface.Config, dict]):
        """ Add a interface
        
        This will also change the manager config.interface_map 
        
        Args:
            name (str): The name (attribute) of the interface
                
                ::
                
                    parent.add_interface('motor1', motor1)
                    parent.motor1 is motor1
            
            interface (Interface, Config , dict): A interface object or a config to build the new interface
        
        """    
        if isinstance(interface, dict):
            type_ = interface.get('type',None)
            if type_ is None:
                cls = self.Interface
            else:
                cls = get_class(KINDS.INTERFACE, type_)
            if name is None:
                raise ValueError("name is missing")                                
            interface = cls.new(self, name, **interface)
        elif isinstance(interface, BaseModel):
            if name is None:
                raise ValueError("name is missing") 
            interface = cls.new(self, name, config=interface)
            
        
        if name is None:
            _, name = ksplit(interface.key)
        
        if name in self.config.interface_map and self.config.interface_map[name] != interface.config:
            raise ValueError("Interface already exists inside its parent")
        elif name in self.__dict__:
            raise ValueError(f"attribute {name} is already taken")
        
        self.config.interface_map[name] = interface.config
        setattr(self, name, interface)              

    def remove_interface(self, name):
        """ Remove a interface 
        
        This will remove also the configuration inside the config.interface_map 
        An ValueError is raised if the interface is not in the parent or if it has been defined 
        statisticaly on the class.
        
        Args:
            name (str) : the reference (the attribute) of the interface
        """
        
        try:
            getattr(self.__class__, name)
        except AttributeError:
            if not name in self.config.interface_map:
                raise ValueError("Unknown interface {name!r}")
            
            del self.config.interface_map[name]
            try:
                del self.__dict__[name]
            except KeyError:
                pass    
        else:
            raise ValueError("interface {name!r} was defined inside the class")    
        
                

interfaceproperty = BaseInterface.prop
    



from pydevmgr_core import * 


class D(BaseDevice):
    my_node = StaticNode.prop('my_node', value=99)
    @NodeAlias.prop('my_alias', nodes=['my_node'])
    def my_dev_alias(self, my_node_value):
        return my_node_value * 10

class D2(BaseDevice):
    d2_node = StaticNode.prop('d2_node', value=78)
    
class I(BaseInterface):
    _devices = []
    
    @BaseNode.prop('all_my_node')
    def all_my_node(self):
        return download([d.my_node  for d in  self._devices])
    
    static = StaticNode.prop('static', value=56)
    

class M(BaseManager):
    d = D.prop('d')
    @NodeAlias.prop('my_alias', nodes=[('d','my_node')])
    def my_alias(self, my_node_value):
        return my_node_value * 10
    
    @I.prop('stat')
    def stat(self, stat):        
        stat._devices = list(self.devices)

class R(BaseRpc):
    class Config(BaseRpc.Config):
        offset: float = 0.0
    def fcall(self):
        print(f"RPC called with offset {self.config.offset}")
        return self.config.offset

    
y = """
node_map: 
    my_dynamic_alias: 
        type: Formula1
        node: my_alias 
        formula: x*10
    my_dynamic_alias2:
        type: Formula
        nodes: [['d','my_node']]
        formula: x*10
"""

    
def test_main():
    global y 
    #m = M(node_map=dynamic_nodes)
    m = M(config=io.load_yaml(y))
    assert m.my_dynamic_alias.get() == 99*10*10   
    assert m.my_dynamic_alias2.get() == 99*10 
    assert m.stat.all_my_node.get() == [99]
    assert D.new(m, None, m.config.device_map['d']).config is m.d.config
    
    m.add_device(None,  D2('d2'))
    assert m.d2.d2_node.get() == 78
    assert m.d2 in list(m.devices)
    
    m._clear_all()
    #assert m.d2 in list(m.devices)
    assert m.d2.d2_node.get() == 78
    
    m.d2.add_node('n10', StaticNode(value=67) )
    assert m.d2.n10.get() == 67
    
    m.d2.add_interface('i10', I())
    m.d2.i10.static.get() == 56
    
    m.d2.i10.add_rpc('move', R())
    assert m.d2.i10.move.call() == 0 
    m.d2.i10.config.rpc_map['move'].offset = 90
    assert m.d2.i10.move.call() == 90
    
    m.add_device( 'd3', {'type':'Base', 'node_map':{'a':{'type':'Static','value':84}}})
    
    
    assert m.d3.a.get() == 84
    d3 = m.d3
    assert d3 in m.devices
    m.remove_device('d3')
    assert d3 not in m.devices
    
    
if __name__ == "__main__":
    test_main()   



from pydevmgr_core import * 
from pydevmgr_core import io 
import os

# add the pkg resource to the cfg path, it contains test_node.yml, test_device.yml 
io.append_cfgpath( io.pkg_res.directory)


nodes = {
    'test1':{'cfgfile':'test_node.yml'},
    'test2':{'cfgfile':'test_node.yml', 'value':0}, 
    'test3': 'test_node.yml',
}


def test_main():

     i = BaseInterface( node_map = nodes)

     assert i.test1.config.value == 9
     assert i.test2.config.value == 0
     assert i.test3.config.value == 9
    
     io.load_config('test_device.yml')

     d = open_object('test_device.yml')
     d = BaseDevice.from_cfgfile('test_device.yml')
     d.i1.n1.get() == 9
     d.i1.n2.get() == 0
     d.n3.get() == 9
  


if __name__ == "__main__":
     test_main()   



from . import test_cfgfile
from . import test_datalink
from . import test_dev_parser
from . import test_device
from . import test_device_config
from . import test_extra_node
from . import test_manager
from . import test_parser

def test_main():
    test_cfgfile.test_main()
    test_datalink.test_main()
    test_dev_parser.test_main()
    test_device.test_main()
    test_device_config.test_main()
    test_extra_node.test_main()
    test_manager.test_main()
    test_parser.test_main()
    
if __name__ == "__main__":
    test_main()



from pydevmgr_core import * 
import time

N = 1000

@to_node_class
def Toto1(scale=1.0, mean=0.0):
    return [scale, mean]
        
class Toto2(BaseNode):
    class Config(BaseNode.Config):
        scale: float = 1.0
        mean: float = 0.0
    def fget(self):
        return [self.config.scale, self.config.mean]

t1 = Toto1('')
t2 = Toto2('')
        
tic = time.time()
for i in range(N):
    t1.get()
toc = time.time()

print("Node Decorator ", (toc-tic)*1e6/N)

tic = time.time()
for i in range(N):
    t2.get()
toc = time.time()

print("Node Class", (toc-tic)*1e6/N)


@record_class
@to_parser_class
def Linear1(x, a=1.0, b=0.0, type="Linear"):
    return a*x+b
    
class Linear2(BaseParser):
    class Config(BaseParser.Config):
        a: float = 1.0
        b: float = 0.0
    @staticmethod
    def parse(x, config):
        return config.a*x + config.b

l1 = get_parser_class("Linear")()
l2 = Linear2()

tic = time.time()
for i in range(N):
    l1(10)
toc = time.time()

print("Parser Decorator", (toc-tic)*1e6/N)

tic = time.time()
for i in range(N):
    l2(10)
toc = time.time()

print("Parser Class", (toc-tic)*1e6/N)



import yaml
import os

from pydantic import BaseModel
from typing import Tuple, Optional, List, Dict, Optional, Callable, Any
from ._yaml_loader import PydevmgrLoader
from ._pkg_resource import PkgResource 

class IOConfig(BaseModel):
    cfgpath : str = 'CFGPATH'
    yaml_ext : Optional[List[str]] = ('.yml', '.yaml')
    YamlLoader = PydevmgrLoader
    
ioconfig = IOConfig()
pkg_res = PkgResource('pydevmgr_core', 'resources')    



def load_config(file_name: str, ioconfig: IOConfig = ioconfig) -> Dict:
    """ Load a configuration file into a dictionary
    
    The file name is given, it should be inside one of the path defined by the $CFGPATH 
    environment variable (or the one defined inside ioconfig.cfgpath). 
    
    Alternatively the path can be an absolute '/' path 

    For now only yaml configuration files are supported 

    Args:
        file_name (str): config file name. Should be inside one of the path defined by the $CFGPATH 
        environment variable. Alternatively the file_name can be an abosolute path
        ioconfig (optional): a :class:`IOConfig`        

    Returns:
        cfg (dict): config dictionary
    """
    return read_config(find_config(file_name, ioconfig = ioconfig), ioconfig = ioconfig)

def read_config(file: str, ioconfig: IOConfig =ioconfig) -> Dict:    
    """ Read the given configuration (yaml) file 
    
    Args:
        file (str): real file path, shall be a yaml file. 
    """
    with open(file) as f:
        return yaml.load(f.read(), Loader=ioconfig.YamlLoader)

def load_yaml(input: str, ioconfig: IOConfig =ioconfig) -> Any:
    return yaml.load(input, Loader=ioconfig.YamlLoader)

def find_config(file_name, ioconfig: IOConfig =ioconfig):
    """ find a config file and return its absolute path 
    
    Args:
        file_name (str): config file name. Should be inside one of the path defined by the $CFGPATH 
        environment variable. Alternatively the file_name can be an abosolute path
        ioconfig (optional): a :class:`IOConfig`        
    """
    path_list = os.environ.get(ioconfig.cfgpath, '.').split(':')
    for directory in path_list[::-1]:
        path = os.path.join(directory, file_name)
        if os.path.exists(path):
            return  path
    raise ValueError('coud not find config file %r in any of %s'%(file_name, path_list))

# overwrite the PydevmgrLoader.find_config method 
PydevmgrLoader.find_config = staticmethod(find_config)

    
def explore_config(filter: Optional[Callable] =None, ioconfig: IOConfig = ioconfig):
    """ Iterator on all config files found inside directories of the $CFGPATH environment variable 
    
    The iterator returns 

    Args:
        filter (None, callable, optional): if given it will receive the content for each file
               to filter
        ioconfig (optional): a :class:`IOConfig`               
    """
        
    path_list = os.environ.get(ioconfig.cfgpath, '.').split(':')
    found = set()
    
    for root in path_list[::-1]:
        for r, d, files in os.walk(root):
            for file in files:
                body, ext = os.path.splitext(file)
                if ext in ioconfig.yaml_ext:
                    p = os.path.relpath( os.path.join(r,file), root )
                    if not p in found:
                        if filter:
                            if filter(read_config(os.path.join(r,file))):
                                yield  p, root
                        else:
                            yield  p, root                                    
                    found.add(p)



def append_cfgpath( path: str, ioconfig: IOConfig = ioconfig):
    """ Add a new path to the $CFGPATH environment variable 

    Args:
        path (str): new config file directory 
        ioconfig (optional): a :class:`IOConfig` 
    """
    envpath = os.environ.get(ioconfig.cfgpath, '').split(':')
    envpath.append(path)
    os.environ[ioconfig.cfgpath] = ":".join(envpath)



